var DocumentScheme = require("./DocumentScheme.js");

var billon = {

    // Attributes
    SoapService: require("./SoapService.js"),

    // Functions
    uniqueId: require('node-uuid').v1
};

billon.Shop = function() {
    this.Merchant = false;
    this.App24 = false;

    this.User = function(cb) {
        cb(false);
    };

    this.Products = function(cb) {
        cb(false);
    };

    this.Product = function(cb) {
        cb(false);
    };

    this.ProductBillon = function(cb) {
        cb(false);
    };

    this.ProductDelivery = function(cb) {
        cb(false);
    };

    this.ProductDownload = function(cb) {
        cb(false);
    };

    this.PaymentTitle = function(cb) {
        cb(false);
    };

    this.DocumentParse = function(cb) {
        cb(false);
    };

    this.OnFinishedPayment = function() {
        console.log("Finished some payment.");
    };

};

// HELPERS
var isFunction = function(func) {
    if (typeof func !== "function") {
        return false;
    }

    return true;
};

var deliveryOptions = {
    "NOT_NEEDED": 0,
    "OPTIONAL": 1,
    "OBLIGATORY": 2
};

var uids = [];

var taskTracker = {};

// ERRORS
var errs = {
    ERR_INVALID_BODY: {
        err: "ERR_INVALID_BODY",
        description: "Missing body."
    },
    ERR_INVALID_QUERY: {
        err: "ERR_INVALID_QUERY",
        description: "Missing query."
    },
    ERR_INVALID_QUERY_DATA: {
        err: "ERR_INVALID_QUERY_DATA",
        description: "Data must be an object."
    },
    ERR_INVALID_PRODUCTS: {
        err: "ERR_INVALID_PRODUCTS",
        description: "Data.products must be an object."
    },
    ERR_INVALID_USER: {
        err: "ERR_INVALID_USER",
        description: "Data.user must be an object."
    },
    ERR_INVALID_WUI: {
        err: "ERR_INVALID_WUI",
        description: "Data.wui must be an object."
    },
    ERR_INVALID_USER_USERNAME: {
        err: "ERR_INVALID_USER_USERNAME",
        description: "Data.user.userName must be a string with minimum 4 characters length."
    },
    ERR_INVALID_TASKID: {
        err: "ERR_INVALID_TASKID",
        description: "Invalid taskId sent."
    },
    ERR_INVALID_UID: {
        err: "ERR_INVALID_UID",
        description: "Invalid uid sent."
    },
    ERR_TASK_DOES_NOT_EXIST: {
        err: "ERR_TASK_DOES_NOT_EXISTS",
        description: "Task does not exists."
    },
    ERR_CANNOT_DOWNLOAD: {
        err: "ERR_CANNOT_DOWNLOAD",
        description: "Cannot download file."
    }
};

billon.Shop.prototype = {
    merchant: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.merchant parameter must be a function.");
            return false;
        }

        this.Merchant = func;
        return this;
    },
    app24: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.app24 parameter must be a function.");
            return false;
        }

        this.App24 = func;
        return this;
    },
    user: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.user parameter must be a function.");
            return false;
        }

        this.User = func;
        return this;
    },
    products: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.products parameter must be a function.");
            return false;
        }

        this.Products = func;
        return this;
    },
    product: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.product parameter must be a function.");
            return false;
        }

        this.Product = func;
        return this;
    },
    productBillon: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.productBillon parameter must be a function.");
            return false;
        }

        this.ProductBillon = func;
        return this;
    },
    productDelivery: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.productDelivery parameter must be a function.");
            return false;
        }

        this.ProductDelivery = func;
        return this;
    },
    productDownload: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.productDownload parameter must be a function.");
            return false;
        }

        this.ProductDownload = func;
        return this;
    },
    documentParse: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.documentParse parameter must be a function.");
            return false;
        }

        this.DocumentParse = func;
        return this;
    },
    paymentTitle: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.paymentTitle parameter must be a function.");
            return false;
        }

        this.PaymentTitle = func;
        return this;
    },
    onFinishedPayment: function(func) {
        if (!isFunction(func)) {
            console.log("Shop.onFinishedPayment parameter must be a function.");
            return false;
        }

        this.OnFinishedPayment = func;
        return this;
    },
    /*
    donateInitiation: function(req, res) {
        var self = this;
        if (typeof req.body !== "object") {
            return res.send(errs.ERR_INVALID_QUERY);
        }

        var requestData = req.body;

        if (typeof requestData.user !== "object") {
            return res.send(errs.ERR_INVALID_USER);
        }

        if (typeof requestData.amount !== "number") {
            return res.send(errs.ERR_INVALID_ITEMS_ID);
        }

        if (typeof requestData.user.userName !== "string" || requestData.user.userName.length < 4) {
            return res.send(errs.ERR_INVALID_USER_USERNAME);
        }

        var paymentData = {
            user: requestData.user,
            amount: requestData.amount
        };

        var requestInvoiceData = "NOT_NEEDED";
        var requestDeliveryData = "NOT_NEEDED";
        var wui = self.App24(requestData);

        var document = DocumentScheme();
        var parsedDocument = this.DocumentParse(document, requestData);
        if (parsedDocument !== false) {
            document = parsedDocument;
        }

        var data = {
            WUI: wui,
            params: {
                type: "NORMAL",
                amount: {
                    amount: paymentData.amount,
                    currency: "PLN"
                },
                requestData: {
                    username: paymentData.user.userName,
                    uniqueTransferId: billon.uniqueId(),
                    title: self.PaymentTitle(),
                    requestInvoiceData: requestInvoiceData,
                    requestDeliveryData: requestDeliveryData,
                    description: document ? JSON.stringify(document) : null
                }
            }
        };

        billon.SoapService.call.requestPaymentInitiation(data, function(requestResponse) {
            var copy = _.clone(requestResponse);
            if (requestResponse.requestStatus === "SUCCESS") {
                var uid = billon.uniqueId();
                taskTracker[uid] = {
                    taskId: requestResponse.taskId,
                    WUI: wui,
                    paymentData: paymentData
                };
                var interval = setInterval(function() {
                    billon.SoapService.call.getTaskStatus({
                        WUI: wui,
                        params: {
                            taskId: requestResponse.taskId
                        }
                    }, function(taskResponse) {
                        if (taskResponse.requestStatus !== "SUCCESS") {
                            setTimeout(function() {
                                delete taskTracker[uid];
                            }, 30000);
                            return clearInterval(interval);
                        }

                        if (taskResponse.status.split("_")[0] === "FINISHED" || taskResponse.progressPercent === 100 || taskResponse.progressPercent === "100") {
                            clearInterval(interval);
                            setTimeout(function() {
                                delete taskTracker[uid];
                            }, 30000);
                        }

                        if (taskResponse.status === "FINISHED_OK") {
                            self.OnFinishedPayment(paymentData);
                        }

                    });
                }, 5000);
                copy.taskId = uid;
            }

            res.json(copy);
        });
    },
    */
    paymentInitiation: function(req, res) {
        var self = this;
        if (typeof req.body !== "object") {
            return res.send(errs.ERR_INVALID_BODY);
        }

        var requestData = req.body,
            paymentData = {};

        self.User(requestData, function(user) {
            if (typeof user !== "object") {
                return res.send(errs.ERR_INVALID_USER);
            }

            if (typeof user.userName !== "string" || user.userName.length < 4) {
                return res.send(errs.ERR_INVALID_USER_USERNAME);
            }

            paymentData.user = user;

            self.Products(requestData, function(products) {
                if (typeof products !== "object") {
                    return res.send(errs.ERR_INVALID_PRODUCTS);
                }

                paymentData.products = products;

                var document = DocumentScheme();

                self.DocumentParse(requestData, document, function(parsedDocument) {
                    if (parsedDocument !== false) {
                        document = parsedDocument;
                    }

                    var requestInvoiceData = "NOT_NEEDED";
                    var requestDeliveryData = "NOT_NEEDED";

                    // Wczytanie produktów.
                    var itemsLoaded = 0;
                    paymentData.items = [];
                    var productsLength = paymentData.products.length;
                    var onItemReceived = function(item) {
                        paymentData.items.push(item);
                        self.ProductBillon(requestData, item, function(currentItem) {
                            if (typeof currentItem !== "object") {
                                console.log("Empty item.");
                                return;
                            }
                            currentItem.unit_price_net = parseInt(currentItem.unit_price_net * 1000);
                            currentItem.amount_net = currentItem.unit_price_net * currentItem.quantity;
                            document.sales_document_data.total_amount_net += currentItem.amount_net;
                            currentItem.amount_gross = currentItem.amount_gross * 1000;
                            document.sales_document_data.total_amount_gross += currentItem.amount_gross * currentItem.quantity;
                            // Sprawdzenie czy potrzeba danych do faktury lub dostawy.
                            if (deliveryOptions[currentItem.requestInvoiceData] > deliveryOptions[requestInvoiceData]) {
                                requestInvoiceData = currentItem.requestInvoiceData;
                            }

                            if (deliveryOptions[currentItem.requestDeliveryData] > deliveryOptions[requestDeliveryData]) {
                                requestDeliveryData = currentItem.requestDeliveryData;
                            }

                            // Pomnożenie vat'u do danych dokumentu.
                            currentItem.vat.amount *= 1000;

                            // Dodanie produktu do listy.
                            document.sales_document_data.items.push(currentItem);

                            itemsLoaded++;
                            if (itemsLoaded === productsLength) {
                                onItemsLoaded();
                            }
                        });
                    };

                    for (var i = 0; i < productsLength; i++) {
                        var itemId = paymentData.products[i];
                        self.Product(requestData, itemId, onItemReceived);
                    }

                    function onItemsLoaded() {

                        // Wczytanie opcji dostawy.
                        self.ProductDelivery(requestData, document.sales_document_data.items, function(delivery) {
                            document.sales_document_data.delivery_option = delivery;

                            // Wczytanie danych sprzedawcy.
                            self.Merchant(requestData, document.sales_document_data.items, document.sales_document_data.delivery_option, function(merchant) {
                                document.seller_data = merchant;

                                // Ustalenie daty.
                                document.sales_document_data.sales_document_date = new Date().getTime();
                                document.sales_document_data.trade_date = new Date().getTime();

                                self.App24(requestData, document.sales_document_data.items, document.sales_document_data.delivery_option, function(wui) {
                                    if (typeof wui !== "object") {
                                        return res.send(errs.ERR_INVALID_WUI);
                                    }
                                    self.PaymentTitle(requestData, document.sales_document_data.items, document.sales_document_data.delivery_option, function(title) {

                                        var keyCodeArr = [];
                                        var counter = 0;
                                        var keyCodeItems = _.filter(paymentData.items, function(item) {
                                            return item.KeyCodes;
                                        });
                                        var keyCodeItemsLength = keyCodeItems.length;

                                        var errorCalled = false;
                                        var onKeyCodeDownload = function(download) {
                                            if (errorCalled) {
                                                return;
                                            }

                                            if (download.err) {
                                                errorCalled = true;
                                                return res.json(errs.ERR_CANNOT_DOWNLOAD);
                                            }

                                            if (!download || download.length === 0) {
                                                return res.json(errs.ERR_CANNOT_DOWNLOAD);
                                            }

                                            // TODO: Obsłużyć wiele kodów jako 1 produkt.
                                            KeyCode.update({
                                                id: download[0].id
                                            }, {
                                                Status: "blocked"
                                            }).exec(function(err, updated) {
                                                if (err) {
                                                    console.log(err);
                                                    return;
                                                }

                                                if (typeof updated !== "undefined") {
                                                    keyCodeArr.push(updated[0]);
                                                }

                                                counter++;
                                                if (keyCodeItemsLength === counter) {
                                                    onDownloadedFinished();
                                                }
                                            });
                                        };

                                        if (keyCodeItemsLength === 0) {
                                            return self.ProductDownload(requestData, paymentData.items, delivery, document, onDownloadedFinished);
                                        }

                                        for (var i = 0; i < keyCodeItemsLength; i++) {
                                            self.ProductDownload(requestData, paymentData.items, delivery, document, onKeyCodeDownload);
                                        }

                                        function onDownloadedFinished(downloadArr) {
                                            var data = {
                                                WUI: wui,
                                                params: {
                                                    type: "NORMAL",
                                                    amount: {
                                                        amount: document.sales_document_data.total_amount_gross,
                                                        currency: "PLN"
                                                    },
                                                    requestData: {
                                                        username: paymentData.user.userName,
                                                        uniqueTransferId: billon.uniqueId(),
                                                        title: title,
                                                        requestInvoiceData: requestInvoiceData,
                                                        requestDeliveryData: requestDeliveryData,
                                                        description: JSON.stringify(document)
                                                    }
                                                }
                                            };

                                            billon.SoapService.call.requestPaymentInitiation(data, function(requestResponse) {
                                                var onUpdated = function(err, updated) {
                                                    if (err) {
                                                        console.log(err);
                                                    }
                                                };

                                                var unBlockKeyCodes = function() {
                                                    for (var i = 0; i < keyCodeArr.length; i++) {
                                                        var keyCode = keyCodeArr[i];
                                                        KeyCode.update({
                                                            id: keyCode.id
                                                        }, {
                                                            Status: "available"
                                                        }).exec(onUpdated);
                                                    }
                                                };

                                                if (requestResponse.err || requestResponse.requestStatus !== "SUCCESS") {
                                                    unBlockKeyCodes();
                                                }

                                                var copy = _.clone(requestResponse);
                                                if (requestResponse.requestStatus === "SUCCESS") {

                                                    var uid = billon.uniqueId();

                                                    taskTracker[uid] = {
                                                        requestData: requestData,
                                                        taskId: requestResponse.taskId,
                                                        WUI: wui,
                                                        items: paymentData.items,
                                                        delivery: document.sales_document_data.delivery_option,
                                                        document: document,
                                                        download: downloadArr,
                                                        keyCodes: keyCodeArr
                                                    };

                                                    var interval = setInterval(function() {
                                                        billon.SoapService.call.getTaskStatus({
                                                            WUI: wui,
                                                            params: {
                                                                taskId: requestResponse.taskId
                                                            }
                                                        }, function(taskResponse) {
                                                            if (taskResponse.requestStatus !== "SUCCESS") {
                                                                clearInterval(interval);
                                                                setTimeout(function() {
                                                                    delete taskTracker[uid];
                                                                }, 30000);

                                                                unBlockKeyCodes();
                                                                return;
                                                            }

                                                            var onUpdatedKey = function(err, updated) {
                                                                if (err) {
                                                                    console.log(err);
                                                                }
                                                            };

                                                            var keyCodeUsed;

                                                            if (taskResponse.status.split("_")[0] === "FINISHED" || taskResponse.progressPercent === 100 || taskResponse.progressPercent === "100") {
                                                                clearInterval(interval);
                                                                setTimeout(function() {
                                                                    delete taskTracker[uid];
                                                                }, 30000);

                                                                if (taskResponse.status !== "FINISHED_OK") {
                                                                    for (var j = 0; j < keyCodeArr.length; j++) {
                                                                        keyCodeUsed = keyCodeArr[j];
                                                                        KeyCode.update({
                                                                            id: keyCodeUsed.id
                                                                        }, {
                                                                            Status: "available"
                                                                        }).exec(onUpdatedKey);
                                                                    }
                                                                }

                                                            }

                                                            if (taskResponse.status === "FINISHED_OK") {
                                                                for (var k = 0; k < keyCodeArr.length; k++) {
                                                                    keyCodeUsed = keyCodeArr[k];
                                                                    console.log("keyCodeUsed");
                                                                    console.log(keyCodeUsed);
                                                                    KeyCode.update({
                                                                        id: keyCodeUsed.id
                                                                    }, {
                                                                        Status: "sold"
                                                                    }).exec(onUpdatedKey);
                                                                }

                                                                self.OnFinishedPayment(document.sales_document_data.items, document.sales_document_data.delivery_option, document, requestData);
                                                            }

                                                        });
                                                    }, 5000);
                                                    copy.taskId = uid;
                                                }

                                                res.json(copy);

                                            });
                                        }
                                    });
                                });
                            });
                        });
                    }
                });
            });
        });
    },
    paymentStatus: function(uid, cb) {
        if (typeof uid !== "string") {
            return cb(errs.ERR_INVALID_UID);
        }

        var self = this;
        var taskData = taskTracker[uid];
        if (typeof taskData === "undefined") {
            return cb(errs.ERR_TASK_DOES_NOT_EXIST);
        }

        billon.SoapService.call.getTaskStatus({
            WUI: taskData.WUI,
            params: {
                taskId: taskData.taskId
            }
        }, function(taskResponse) {

            if (typeof taskResponse.status === "string" && taskResponse.status === "FINISHED_OK") {
                var publicCodes = [];
                for (var i = 0; i < taskData.keyCodes.length; i++) {
                    publicCodes.push(taskData.keyCodes[i].Value);
                }

                taskResponse.agent = {
                    download: taskData.download,
                    keyCodes: publicCodes
                };

                if (cb) {
                    cb(taskResponse);
                }
            } else if (cb) {
                cb(taskResponse);
            }
        });
    }
};

module.exports = billon;
